import asyncio
import os
import platform
import shlex
import sys
from typing import Any, List, Optional, Union

import tomodachi
import tomodachi.importer
from tomodachi import logging
from tomodachi.helpers.build_time import get_time_since_build
from tomodachi.helpers.colors import NO_COLOR
from tomodachi.helpers.execution_context import get_execution_context
from tomodachi.importer import ServiceImporter

TOMODACHI_ASCII = """
▀█▀ █▀█ █▀▄▀█ █▀█ █▀▄ ▄▀█ █▀▀ █ █ █
 █  █▄█ █ ▀ █ █▄█ █▄▀ █▀█ █▄▄ █▀█ █  友達
""".strip()

if NO_COLOR or any(
    [
        os.environ.get("TOMODACHI_BANNER_NO_COLOR", "").lower() in ("1", "true"),
        os.environ.get("TOMODACHI_BANNER_NOCOLOR", "").lower() in ("1", "true"),
    ]
):
    from tomodachi.helpers.colors import COLOR_0 as COLOR
    from tomodachi.helpers.colors import COLOR_RESET_0 as COLOR_RESET
    from tomodachi.helpers.colors import COLOR_STYLE_0 as COLOR_STYLE
else:
    from tomodachi.helpers.colors import COLOR, COLOR_RESET, COLOR_STYLE


def render_banner(
    service_files: Union[List, set],
    loop: Optional[asyncio.AbstractEventLoop] = None,
    process_id: Optional[int] = None,
    event_loop_alias: Optional[str] = None,
    event_loop_version: Optional[str] = None,
    cwd: Optional[str] = None,
    watcher_enabled: Optional[bool] = None,
    use_execution_context: bool = True,
) -> None:
    output = []

    context = get_execution_context() if use_execution_context else {}

    if process_id is None:
        process_id = context.get("process_id", os.getpid()) or os.getpid()

    if cwd is None:
        cwd = context.get("cwd", os.getcwd()) or os.getcwd()

    if loop is None:
        if sys.version_info.major == 3 and sys.version_info.minor < 10:
            loop = asyncio.get_event_loop()
        else:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()

    if event_loop_alias is None:
        event_loop_alias = context.get("event_loop")
    if event_loop_version is None:
        event_loop_version = context.get("event_loop_version")
    if event_loop_alias is None:
        event_loop_alias = ""
        try:
            if "uvloop." in str(loop.__class__):
                event_loop_alias = "uvloop"
                import uvloop  # noqa  # isort:skip

                if event_loop_version is None:
                    event_loop_version = str(getattr(uvloop, "__version__", ""))
            elif "asyncio." in str(loop.__class__):
                event_loop_alias = "asyncio"
            else:
                event_loop_alias = "{}.{}".format(loop.__class__.__module__, loop.__class__.__name__)
        except Exception:
            event_loop_alias = str(loop)
    if event_loop_version is None:
        event_loop_version = ""

    if watcher_enabled is None:
        watcher_enabled = context.get("watcher_enabled", False)

    event_loop_setting = str(tomodachi.context("loop.setting") or "")

    output.append("")
    output.append(
        COLOR_RESET + COLOR.LIGHTBLACK_EX + COLOR_STYLE.BRIGHT + COLOR_STYLE.DIM + TOMODACHI_ASCII + COLOR_RESET
    )
    output.append("")

    arg0 = sys.argv[0]
    if arg0.endswith("/__main__.py"):
        for path in sys.path:
            if arg0.startswith(path + "/"):
                arg0 = arg0.split(path + "/", 1)[-1].rsplit("/__main__.py", 1)[0]
                break

        arg0 = arg0.replace("/", ".")
        process_cmd = ["python", "-m", arg0] + sys.argv[1:]
    else:
        process_cmd = [arg0.rsplit("/", 1)[-1]] + sys.argv[1:]

    actual_file_paths = []
    potential_file_paths = set()

    importer_logger_disabled_value = logging.get_logger("tomodachi.importer")._logger.disabled
    logging.get_logger("tomodachi.importer")._logger.disabled = True

    for file_path in service_files:
        file_path_ = file_path
        potential_file_paths.add(file_path)

        try:

            async def _import_module_without_logger() -> Any:
                return ServiceImporter.import_service_file(file_path, no_exec=True)

            service_import = loop.run_until_complete(_import_module_without_logger())
            if service_import and service_import.__file__:
                file_path_ = service_import.__file__
                potential_file_paths.add(file_path_)
                if file_path_.endswith(".pyc"):
                    file_path_ = file_path_[:-1]
                    potential_file_paths.add(file_path_)

                if cwd.rstrip("/") and file_path_.startswith(cwd.rstrip("/") + "/"):
                    file_path_ = file_path_.split(cwd.rstrip("/") + "/", 1)[-1]
                    potential_file_paths.add(file_path_)
                    file_path_ = "./" + file_path_
                    potential_file_paths.add(file_path_)
        except tomodachi.importer.ServicePackageError:
            pass
        except (SyntaxError, IndentationError):
            pass
        except Exception:
            pass

        if file_path_ not in actual_file_paths:
            actual_file_paths.append(file_path_)

    logging.get_logger("tomodachi.importer")._logger.disabled = importer_logger_disabled_value

    for file_path in potential_file_paths:
        if file_path in process_cmd:
            process_cmd = [arg for arg in process_cmd if arg != file_path]
            if "<...>" not in process_cmd:
                process_cmd.append("<...>")

    for value in ("-c", "--config"):
        if value in process_cmd:
            config_file_path = (process_cmd[process_cmd.index(value, 0) :] + ["..."])[1]
            if config_file_path.endswith(".json"):
                process_cmd = [arg if arg != config_file_path else "<...>" for arg in process_cmd]

    process_cmd = [arg for arg in process_cmd if arg.strip()]
    process_cmd_str: str
    while True:
        process_cmd_str = shlex.join(process_cmd).replace("'<...>'", "<...>")
        if len(process_cmd_str) <= 60:
            break
        process_cmd = [arg for arg in process_cmd if arg.strip() and arg != "..."][:-1] + ["..."]

    os_name = platform.system()
    if os_name == "Darwin":
        os_name = "macOS"

    machine = platform.machine() or platform.processor()
    node_name = platform.node() or os.environ.get("HOSTNAME", "") or "..."

    LABEL = f"{COLOR_RESET}{COLOR.WHITE}{COLOR_STYLE.BRIGHT}{COLOR_STYLE.DIM}"
    DELIMITER = f"{COLOR_RESET}{COLOR_STYLE.DIM}⇢{COLOR_RESET}"
    TEXT_NORMAL = f"{COLOR_RESET}{COLOR.WHITE}{COLOR_STYLE.DIM}"
    TEXT_HIGHLIGHT = f"{COLOR_RESET}{COLOR.WHITE}"
    NOTICE_TEXT_TOPIC_WATCHER = f"{COLOR_RESET}{COLOR.LIGHTCYAN_EX}{COLOR_STYLE.DIM}{COLOR_STYLE.BRIGHT}"
    NOTICE_TEXT_TOPIC_EXIT = f"{COLOR_RESET}{COLOR.LIGHTYELLOW_EX}{COLOR_STYLE.DIM}{COLOR_STYLE.BRIGHT}"
    PID_HIGHLIGHT = f"{COLOR_RESET}{COLOR.GREEN}{COLOR_STYLE.BRIGHT}{COLOR_STYLE.DIM}"

    pid_str = f"[pid: {process_id}]"
    pid_str = f"{pid_str:<14}"
    pid_str = (
        LABEL
        + "process "
        + pid_str[0:1]
        + PID_HIGHLIGHT
        + pid_str[1 : 6 + len(str(process_id))]
        + LABEL
        + pid_str[6 + len(str(process_id)) :]
    )

    output.append(f"{pid_str} {DELIMITER} {TEXT_HIGHLIGHT}$ {process_cmd_str}{COLOR_RESET}")

    for file_num, file_path in enumerate(actual_file_paths, 1):
        file_path_ = file_path
        while True:
            if len(file_path_) <= 60:
                break
            file_path_ = ".../" + "/".join(file_path_.replace(".../", "").split("/")[1:])
            if file_path_.startswith(".../") and file_path_.count("/") == 1:
                file_path_ = file_path_[4:]
                break

        if len(actual_file_paths) == 1:
            file_num_ = ""
        else:
            file_num_ = f"[{file_num}]"
        output.append(f"{LABEL}service file {file_num_:4s}      {DELIMITER} {TEXT_HIGHLIGHT}{file_path_}{COLOR_RESET}")

    time_since_tomodachi_build = get_time_since_build()

    venv_prompt = ""
    venv_path = ""
    home_path = os.path.expanduser("~").rstrip("/")
    python_exec_prefix = sys.exec_prefix
    python_path = sys.executable
    venv_environ = os.environ.get("VIRTUAL_ENV", "")

    if "venv" in python_exec_prefix or "virtualenv" in python_exec_prefix or venv_environ:
        try:
            venv_path = python_exec_prefix
            if cwd.rstrip("/") and venv_path.startswith(cwd.rstrip("/") + "/"):
                venv_path = "./" + venv_path.split(cwd.rstrip("/") + "/", 1)[-1]
            elif home_path and venv_path.startswith(home_path + "/"):
                venv_path = "~/" + venv_path.split(home_path + "/", 1)[-1]
            venv_config = open(os.path.join(python_exec_prefix, "pyvenv.cfg"), "r").read()
            for line in venv_config.split("\n"):
                if line.startswith("prompt = "):
                    venv_prompt = line.split("prompt = ", 1)[-1].replace('"', "").replace("'", "")
                    break
        except (FileNotFoundError, PermissionError, Exception):
            pass

    if cwd.rstrip("/") and python_path.startswith(cwd.rstrip("/") + "/"):
        python_path = "./" + python_path.split(cwd.rstrip("/") + "/", 1)[-1]
    elif home_path and python_path.startswith(home_path + "/"):
        python_path = "~/" + python_path.split(home_path + "/", 1)[-1]

    poetry_venv = ""
    poetry_is_active = (os.environ.get("POETRY_ACTIVE", "").lower() in ("1", "true") and venv_environ) or (
        "/pypoetry/virtualenvs/" in venv_environ or "/poetry/virtualenvs/" in venv_environ
    )
    if poetry_is_active:
        if cwd.rstrip("/") and venv_environ.startswith(cwd.rstrip("/") + "/"):
            poetry_venv = venv_environ
            if cwd.rstrip("/") and poetry_venv.startswith(cwd.rstrip("/") + "/"):
                poetry_venv = "./" + poetry_venv.split(cwd.rstrip("/") + "/", 1)[-1]
            elif home_path and poetry_venv.startswith(home_path + "/"):
                poetry_venv = "~/" + poetry_venv.split(home_path + "/", 1)[-1]
        elif venv_path.rsplit("/", 1)[-1] in (".venv", "venv"):
            poetry_venv = venv_environ
        else:
            poetry_venv = venv_environ.rsplit("/", 1)[-1]

    if venv_path and python_path.startswith(venv_path + "/"):
        python_path = "$VIRTUAL_ENV/" + python_path.split(venv_path + "/", 1)[-1]
    elif poetry_venv and "/" in poetry_venv and python_path.startswith(poetry_venv + "/"):
        python_path = "$VIRTUAL_ENV/" + python_path.split(poetry_venv + "/", 1)[-1]

    output.append("")

    output.append(
        f"{LABEL}operating system       {DELIMITER} {TEXT_HIGHLIGHT}{os_name} on {machine}{TEXT_NORMAL}"
        + (f" (hostname: {node_name})" if node_name else "")
        + COLOR_RESET
    )

    output.append(
        f"{LABEL}python runtime         {DELIMITER} {TEXT_HIGHLIGHT}{platform.python_implementation()} {platform.python_version()}{TEXT_NORMAL} (build: {platform.python_build()[1]}){COLOR_RESET}"
    )

    output.append(
        f"{LABEL}tomodachi version      {DELIMITER} {TEXT_HIGHLIGHT}{tomodachi.__version__}{TEXT_NORMAL}"
        + (f" ({time_since_tomodachi_build})" if time_since_tomodachi_build else " (local development version)")
        + COLOR_RESET
    )

    if poetry_venv and poetry_venv != venv_path and venv_path.rsplit("/", 1)[-1] not in (".venv", "venv"):
        output.append(
            f"{LABEL}virtualenv             {DELIMITER} {TEXT_HIGHLIGHT}poetry env{TEXT_NORMAL} (name: {poetry_venv})"
            + COLOR_RESET
        )
    elif poetry_venv:
        output.append(
            f"{LABEL}virtualenv path        {DELIMITER} {TEXT_HIGHLIGHT}{poetry_venv}{TEXT_NORMAL} (poetry active)"
            + COLOR_RESET
        )
    elif venv_path:
        if venv_path:
            output.append(
                f"{LABEL}virtualenv path        {DELIMITER} {TEXT_HIGHLIGHT}{venv_path}{TEXT_NORMAL}"
                + (f" (prompt: {venv_prompt})" if venv_prompt and venv_prompt != venv_path.rsplit("/", 1)[-1] else "")
                + COLOR_RESET
            )

    python_path_is_outside_venv = bool(venv_path and "$VIRTUAL_ENV" not in python_path)
    if python_path_is_outside_venv or (
        not venv_path
        and python_path
        not in (
            "/usr/bin/python",
            "/usr/bin/python3",
            "/usr/local/bin/python",
            "/usr/local/bin/python3",
            "/opt/bin/python",
            "/opt/bin/python3",
            "/opt/homebrew/opt/python/bin/python",
            "/opt/homebrew/opt/python/bin/python3",
            "/opt/homebrew/opt/bin/python",
            "/opt/homebrew/opt/bin/python3",
        )
    ):
        output.append(
            f"{LABEL}python executable      {DELIMITER} {TEXT_HIGHLIGHT}{python_path}{TEXT_NORMAL}"
            + (
                f" {COLOR_RESET}{COLOR.LIGHTRED_EX}{COLOR_STYLE.BRIGHT}{COLOR_STYLE.DIM}[outside venv]"
                if python_path_is_outside_venv
                else ""
            )
            + COLOR_RESET
        )

    output.append(
        f"{LABEL}event loop             {DELIMITER} {TEXT_HIGHLIGHT}{event_loop_alias}{TEXT_NORMAL}"
        + (f" (version: {event_loop_version})" if event_loop_version else "")
        + (f" ({event_loop_setting})" if event_loop_setting == "auto" else "")
        + COLOR_RESET
    )

    output.append("")

    if watcher_enabled:
        output.append(
            f"{NOTICE_TEXT_TOPIC_WATCHER}⌁ [watcher] file watcher enabled (code changes auto restart services){COLOR_RESET}"
        )
    output.append(
        f"{NOTICE_TEXT_TOPIC_EXIT}⌁ [notice ] stop running services with <ctrl+c> (graceful teardown){COLOR_RESET}"
    )

    output.append(COLOR_RESET + "")

    print("\n".join(output))



import re
from time import time_ns
from traceback import format_exception
from typing import Any, Awaitable, Callable, Dict, Mapping, NamedTuple, Optional, Tuple, Type, TypeVar, cast

from aiohttp import hdrs, web
from opentelemetry import metrics, trace
from opentelemetry.metrics._internal.instrument import Instrument
from opentelemetry.sdk.metrics import Meter
from opentelemetry.sdk.metrics._internal.instrument import _Histogram, _UpDownCounter
from opentelemetry.sdk.trace import Span
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from opentelemetry.util.http import ExcludeList
from opentelemetry.util.types import AttributeValue

from tomodachi._exception import limit_exception_traceback
from tomodachi.logging import get_logger
from tomodachi.transport.aws_sns_sqs import AWSSNSSQSTransport, MessageAttributesType
from tomodachi.transport.http import get_forwarded_remote_ip

IT = TypeVar("IT", bound=Instrument)


class InstrumentOptions(NamedTuple):
    name: str
    unit: str
    description: str
    attribute_keys: Tuple[str, ...]


class OpenTelemetryTomodachiMiddleware:
    service: Any
    service_context: Dict
    tracer: Optional[trace.Tracer]
    meter: Optional[metrics.Meter]

    duration_histogram_options: InstrumentOptions
    active_tasks_counter_options: InstrumentOptions

    _duration_histogram: Optional[metrics.Histogram] = None
    _active_tasks_counter: Optional[metrics.UpDownCounter] = None

    def __init__(
        self, service: Any, tracer: Optional[trace.Tracer] = None, meter: Optional[metrics.Meter] = None, **kwargs: Any
    ) -> None:
        self.service = service
        self.service_context = getattr(service, "context", {})
        self.tracer = tracer
        self.meter = meter

        if meter:
            self._duration_histogram = self._get_instrument(_Histogram, *self.duration_histogram_options[0:3])
            self._active_tasks_counter = self._get_instrument(_UpDownCounter, *self.active_tasks_counter_options[0:3])

    def _get_instrument(self, type_: Type[IT], name: str, unit: str, description: str) -> Optional[IT]:
        meter = cast(Meter, self.meter)
        (is_registered, instrument_id) = meter._is_instrument_registered(name, type_, unit, description)

        if not is_registered:
            with meter._instrument_ids_lock:
                try:
                    meter._instrument_ids.remove(instrument_id)
                except KeyError:
                    pass

            if type_ is _Histogram:
                return cast(IT, meter.create_histogram(name, unit, description))
            elif type_ is _UpDownCounter:
                return cast(IT, meter.create_up_down_counter(name, unit, description))
            else:
                get_logger("tomodachi.opentelemetry").warning(
                    "unsupported instrument type",
                    instrument_name=name,
                    instrument_unit=unit,
                    instrument_description=description,
                    instrument_type=getattr(type_, "__name__", type(type_).__name__),
                )
                raise Exception("unsupported instrument type")

        with meter._instrument_id_instrument_lock:
            return cast(IT, meter._instrument_id_instrument[instrument_id])

    def increase_active_tasks(
        self,
        attributes: Optional[Mapping[str, AttributeValue]],
        extra_attributes: Optional[Dict[str, AttributeValue]] = None,
        *,
        value: int = 1,
    ) -> None:
        if self._active_tasks_counter and attributes:
            attribute_keys = self.active_tasks_counter_options.attribute_keys
            attributes_ = {k: attributes[k] for k in attribute_keys if k in attributes}
            if extra_attributes:
                attributes_.update(extra_attributes)
            self._active_tasks_counter.add(value, attributes_)

    def decrease_active_tasks(
        self,
        attributes: Optional[Mapping[str, AttributeValue]],
        extra_attributes: Optional[Dict[str, AttributeValue]] = None,
        *,
        value: int = -1,
    ) -> None:
        self.increase_active_tasks(attributes, extra_attributes, value=value)

    def record_duration(
        self,
        start_time: Optional[int],
        end_time: Optional[int],
        attributes: Optional[Mapping[str, AttributeValue]],
        extra_attributes: Optional[Dict[str, AttributeValue]] = None,
    ) -> None:
        if self._duration_histogram and start_time and end_time and attributes:
            attribute_keys = self.duration_histogram_options.attribute_keys
            attributes_ = {k: attributes[k] for k in attribute_keys if k in attributes}
            if extra_attributes:
                attributes_.update(extra_attributes)

            duration_ns = end_time - start_time
            duration = duration_ns / 1e9

            self._duration_histogram.record(duration, attributes_)


@web.middleware
class OpenTelemetryAioHTTPMiddleware(OpenTelemetryTomodachiMiddleware):
    duration_histogram_options = InstrumentOptions(
        "http.server.duration",
        "s",
        "Measures the duration of inbound HTTP requests.",
        (
            "http.route",
            "http.request.method",
            "http.response.status_code",
            "network.protocol.name",
            "network.protocol.version",
            "server.address",
            "server.port",
            "url.scheme",
        ),
    )
    active_tasks_counter_options = InstrumentOptions(
        "http.server.active_requests",
        "{request}",
        "Measures the number of concurrent HTTP requests that are currently in-flight.",
        (
            "http.request.method",
            "server.address",
            "server.port",
            "url.scheme",
        ),
    )

    def __init__(
        self,
        service: Any,
        tracer: Optional[trace.Tracer] = None,
        meter: Optional[metrics.Meter] = None,
        excluded_urls: Optional[ExcludeList] = None,
    ) -> None:
        super().__init__(service, tracer, meter)
        self.excluded_urls = excluded_urls

    def get_route(self, request: web.Request) -> Optional[str]:
        route: Optional[str]
        try:
            route = getattr(request.match_info.route._resource, "_simplified_pattern", None)
        except Exception:
            route = None

        if not route:
            pattern = request.match_info.route.get_info().get("pattern")
            if not pattern:
                return None
            simplified = re.compile(r"^\^?(.+?)\$?$").match(pattern.pattern)
            route = simplified.group(1) if simplified else None

        return route

    def get_attributes(self, request: web.Request) -> Dict[str, AttributeValue]:
        protocol_version = ".".join(map(str, request.version))
        route: Optional[str] = self.get_route(request)

        client_socket_addr = None
        client_socket_port = None
        if request._transport_peername:
            client_socket_addr, client_socket_port = cast(Tuple[str, int], request._transport_peername)

        server_ip = None
        server_port = None
        if request.transport:
            sock = request.transport.get_extra_info("socket")
            if sock:
                sockname = sock.getsockname()
                if sockname:
                    server_ip, server_port = cast(Tuple[str, int], sockname)

        host = request.headers.get(hdrs.HOST)
        port = int(host.split(":")[1]) if host and ":" in host else 80

        attributes: Dict[str, AttributeValue] = {}

        if request.method:
            attributes["http.request.method"] = request.method

        if route:
            attributes["http.route"] = route

        attributes["server.address"] = host.split(":")[0] if host is not None else (server_ip or "127.0.0.1")
        attributes["server.port"] = port if host is not None else (server_port or port)

        attributes["url.scheme"] = request.scheme
        attributes["url.path"] = request.path

        if request.query_string:
            attributes["url.query"] = request.query_string

        user_agent = request.headers.get(hdrs.USER_AGENT)
        if user_agent:
            attributes["user_agent.original"] = user_agent

        attributes["client.address"] = get_forwarded_remote_ip(request)

        if client_socket_addr and client_socket_port:
            attributes["client.socket.address"] = client_socket_addr
            attributes["client.socket.port"] = client_socket_port

        if server_ip:
            attributes["server.socket.address"] = server_ip

        attributes["server.socket.port"] = server_port or port
        attributes["network.protocol.version"] = protocol_version
        attributes["network.protocol.name"] = "http"

        return attributes

    async def __call__(
        self,
        request: web.Request,
        handler: Callable[..., Awaitable[web.Response]],
    ) -> web.Response:
        try:
            return await self.handle(request, handler)
        except web.HTTPException:
            raise
        except Exception as exc:
            get_logger("exception").exception(f"uncaught exception: {str(exc)}")
            raise

    async def handle(
        self,
        request: web.Request,
        handler: Callable[..., Awaitable[web.Response]],
    ) -> web.Response:
        exclude_instrumentation: bool = any(
            [
                getattr(self.service, "_is_instrumented_by_opentelemetry", False) is False,
                self.excluded_urls and self.excluded_urls.url_disabled(request.path),
            ]
        )
        if exclude_instrumentation or not self.tracer:
            return await handler(request)

        start_time = time_ns()

        attributes = self.get_attributes(request)
        route = cast(Optional[str], attributes.get("http.route"))
        ctx = TraceContextTextMapPropagator().extract(carrier=request.headers)
        response: Optional[web.Response] = None

        span = cast(
            Span,
            self.tracer.start_span(
                name=f"{request.method} {route}" if route else request.method,
                kind=trace.SpanKind.SERVER,
                context=ctx,
                attributes=attributes,
                start_time=start_time,
            ),
        )
        with trace.use_span(span, end_on_exit=False):
            try:
                self.increase_active_tasks(attributes)
                response = await handler(request)
            except web.HTTPException as exc:
                response = exc
                if exc.status < 100 or exc.status >= 500:
                    raise
            finally:
                response_status_code: int
                if (
                    response is not None
                    and isinstance(getattr(response, "status", None), int)
                    and response.status >= 100
                    and response.status <= 599
                ):
                    response_status_code = response.status
                    if response.status >= 500:
                        span.set_status(trace.StatusCode.ERROR)
                    if response.status == 499:
                        replaced_status_code = getattr(response, "_replaced_status_code", None)
                        if replaced_status_code:
                            span.set_status(
                                trace.StatusCode.ERROR,
                                "Client closed the connection while the server was still processing the request.",
                            )
                            response_status_code = replaced_status_code
                        else:
                            span.set_status(
                                trace.StatusCode.ERROR,
                                "Client closed the connection before the server could start processing the request.",
                            )

                    caught_exceptions = getattr(response, "_caught_exceptions", None)
                    if caught_exceptions:
                        for exception in caught_exceptions:
                            span.record_exception(
                                exception,
                                {
                                    "exception.stacktrace": "".join(
                                        format_exception(type(exception), exception, exception.__traceback__)
                                    )
                                },
                            )
                else:
                    response_status_code = 500
                    span.set_status(trace.StatusCode.ERROR)
                    response = None

                span.set_attribute("http.response.status_code", response_status_code)

                self.decrease_active_tasks(attributes)
                end_time = span.end_time if span.is_recording() and span.end_time else time_ns()
                self.record_duration(
                    start_time, end_time, attributes, {"http.response.status_code": response_status_code}
                )
                span.end(end_time=end_time)

        if response is None:
            response = web.HTTPInternalServerError()
            response.body = b""
            response.headers[hdrs.CONNECTION] = "close"
            response.force_close()

        if isinstance(response, (web.HTTPException, web.HTTPInternalServerError)):
            raise response

        return response


class OpenTelemetryAWSSQSMiddleware(OpenTelemetryTomodachiMiddleware):
    duration_histogram_options = InstrumentOptions(
        "messaging.amazonsqs.duration",
        "s",
        "Measures the duration of processing a message by the handler function.",
        (
            "messaging.source.name",
            "messaging.source.kind",
            "messaging.destination.name",
            "messaging.destination.kind",
            "code.function",
        ),
    )
    active_tasks_counter_options = InstrumentOptions(
        "messaging.amazonsqs.active_tasks",
        "{message}",
        "Measures the number of concurrent SQS messages that are currently being processed.",
        (
            "messaging.source.name",
            "messaging.source.kind",
            "messaging.destination.name",
            "messaging.destination.kind",
            "code.function",
        ),
    )

    async def __call__(
        self,
        handler: Callable[..., Awaitable[None]],
        *,
        topic: str,
        queue_url: str,
        message_attributes: MessageAttributesType,
        sns_message_id: str,
        sqs_message_id: str,
    ) -> None:
        if getattr(self.service, "_is_instrumented_by_opentelemetry", False) is False or not self.tracer:
            return await handler()

        start_time = time_ns()

        queue_name: str = AWSSNSSQSTransport.get_queue_name_without_prefix(
            AWSSNSSQSTransport.get_queue_name_from_queue_url(queue_url), self.service_context
        )

        attributes: Dict[str, AttributeValue] = {
            "messaging.system": "AmazonSQS",
            "messaging.operation": "process",
            "messaging.source.name": queue_name,
            "messaging.source.kind": "queue",
            "messaging.message.id": sns_message_id or sqs_message_id,
            "messaging.destination.name": topic if topic else queue_name,
            "messaging.destination.kind": "topic" if topic else "queue",
            "code.function": handler.__name__,
        }

        ctx = TraceContextTextMapPropagator().extract(carrier=message_attributes)

        span = cast(
            Span,
            self.tracer.start_span(
                name=f"{topic} process",
                kind=trace.SpanKind.CONSUMER,
                context=ctx,
                attributes=attributes,
                start_time=start_time,
            ),
        )
        with trace.use_span(span, end_on_exit=False):
            function_success: bool = True
            try:
                self.increase_active_tasks(attributes)
                await handler()
                span.set_status(trace.StatusCode.OK)
            except Exception as exc:
                function_success = False
                span.set_status(trace.StatusCode.ERROR)
                limit_exception_traceback(exc, ("tomodachi.transport.aws_sns_sqs", "tomodachi.helpers.middleware"))
                span.record_exception(
                    exc,
                    {"exception.stacktrace": "".join(format_exception(type(exc), exc, exc.__traceback__))},
                    escaped=True,
                )
                raise
            finally:
                self.decrease_active_tasks(attributes)
                end_time = span.end_time if span.is_recording() and span.end_time else time_ns()
                self.record_duration(start_time, end_time, attributes, {"function.success": function_success})
                span.end(end_time=end_time)


class OpenTelemetryAMQPMiddleware(OpenTelemetryTomodachiMiddleware):
    duration_histogram_options = InstrumentOptions(
        "messaging.rabbitmq.duration",
        "s",
        "Measures the duration of processing a message by the handler function.",
        (
            "messaging.destination.name",
            "messaging.rabbitmq.destination.routing_key",
            "code.function",
        ),
    )
    active_tasks_counter_options = InstrumentOptions(
        "messaging.rabbitmq.active_tasks",
        "{message}",
        "Measures the number of concurrent AMQP messages that are currently being processed.",
        (
            "messaging.destination.name",
            "messaging.rabbitmq.destination.routing_key",
            "code.function",
        ),
    )

    async def __call__(
        self,
        handler: Callable[..., Awaitable[None]],
        *,
        routing_key: str,
        exchange_name: str,
        properties: Any,
    ) -> None:
        if getattr(self.service, "_is_instrumented_by_opentelemetry", False) is False or not self.tracer:
            return await handler()

        start_time = time_ns()

        if not exchange_name:
            exchange_name = "amq.topic"

        attributes: Dict[str, AttributeValue] = {
            "messaging.system": "rabbitmq",
            "messaging.operation": "process",
            "messaging.destination.name": exchange_name,
            "messaging.rabbitmq.destination.routing_key": routing_key,
            "code.function": handler.__name__,
        }

        headers = getattr(properties, "headers", {})
        if not isinstance(headers, dict):
            headers = {}
        ctx = TraceContextTextMapPropagator().extract(carrier=headers)

        span = cast(
            Span,
            self.tracer.start_span(
                name=f"{routing_key} process",
                kind=trace.SpanKind.CONSUMER,
                context=ctx,
                attributes=attributes,
                start_time=start_time,
            ),
        )
        with trace.use_span(span, end_on_exit=False):
            function_success: bool = True
            try:
                self.increase_active_tasks(attributes)
                await handler()
                span.set_status(trace.StatusCode.OK)
            except Exception as exc:
                function_success = False
                span.set_status(trace.StatusCode.ERROR)
                limit_exception_traceback(exc, ("tomodachi.transport.amqp", "tomodachi.helpers.middleware"))
                span.record_exception(
                    exc,
                    {"exception.stacktrace": "".join(format_exception(type(exc), exc, exc.__traceback__))},
                    escaped=True,
                )
                raise
            finally:
                self.decrease_active_tasks(attributes)
                end_time = span.end_time if span.is_recording() and span.end_time else time_ns()
                self.record_duration(start_time, end_time, attributes, {"function.success": function_success})
                span.end(end_time=end_time)


class OpenTelemetryScheduleFunctionMiddleware(OpenTelemetryTomodachiMiddleware):
    duration_histogram_options = InstrumentOptions(
        "function.duration", "s", "Measures the duration of running the scheduled handler function.", ("code.function",)
    )
    active_tasks_counter_options = InstrumentOptions(
        "function.active_tasks",
        "{task}",
        "Measures the number of concurrent invocations of the scheduled handler that is currently running.",
        ("code.function",),
    )

    async def __call__(
        self,
        handler: Callable[..., Awaitable[None]],
        *,
        invocation_time: str = "",
        interval: str = "",
    ) -> None:
        if getattr(self.service, "_is_instrumented_by_opentelemetry", False) is False or not self.tracer:
            return await handler()

        start_time = time_ns()

        attributes: Dict[str, AttributeValue] = {
            "code.function": handler.__name__,
            "function.trigger": "timer",
            "function.time": invocation_time,
        }

        if interval:
            attributes["function.interval"] = str(interval)

        span = cast(
            Span,
            self.tracer.start_span(
                name=handler.__name__,
                kind=trace.SpanKind.INTERNAL,
                attributes=attributes,
                start_time=start_time,
            ),
        )
        with trace.use_span(span, end_on_exit=False):
            function_success: bool = True
            try:
                self.increase_active_tasks(attributes)
                await handler()
                span.set_status(trace.StatusCode.OK)
            except Exception as exc:
                function_success = False
                span.set_status(trace.StatusCode.ERROR)
                limit_exception_traceback(exc, ("tomodachi.transport.schedule", "tomodachi.helpers.middleware"))
                span.record_exception(
                    exc,
                    {"exception.stacktrace": "".join(format_exception(type(exc), exc, exc.__traceback__))},
                    escaped=True,
                )
                raise
            finally:
                self.decrease_active_tasks(attributes)
                end_time = span.end_time if span.is_recording() and span.end_time else time_ns()
                self.record_duration(start_time, end_time, attributes, {"function.success": function_success})
                span.end(end_time=end_time)



import asyncio
import inspect
import os
import re
import sys
import types
import uuid
from types import ModuleType, TracebackType
from typing import Any, Callable, Dict, List, Optional, Set, Type, cast

import tomodachi
from tomodachi import CLASS_ATTRIBUTE, logging
from tomodachi._exception import limit_exception_traceback
from tomodachi.helpers.dict import merge_dicts
from tomodachi.helpers.execution_context import set_service, unset_service
from tomodachi.invoker import FUNCTION_ATTRIBUTE, INVOKER_TASK_START_KEYWORD, START_ATTRIBUTE


class ServiceContainer(object):
    def __init__(self, module_import: ModuleType, configuration: Optional[Dict] = None) -> None:
        self.module_import = module_import

        self.file_path = module_import.__file__
        self.module_name = (
            module_import.__name__.rsplit("/", 1)[1] if "/" in module_import.__name__ else module_import.__name__
        ).rsplit(".", 1)[-1]
        self.configuration = configuration
        self.logger = logging.getLogger("tomodachi.lifecycle")

        self._close_waiter: Optional[asyncio.Future] = None
        self.started_waiter: Optional[asyncio.Future] = None

        def catch_uncaught_exceptions(
            type_: Type[BaseException], value: BaseException, traceback: Optional[TracebackType]
        ) -> Any:
            raise value

        sys.excepthook = catch_uncaught_exceptions

    def stop_service(self) -> None:
        tomodachi.get_contextvar("exit_code").set(tomodachi.SERVICE_EXIT_CODE)
        if not self._close_waiter:
            self._close_waiter = asyncio.Future()

        if not self._close_waiter.done():
            self._close_waiter.set_result(None)

    def setup_configuration(self, instance: Any) -> None:
        if not self.configuration:
            return
        for k, v in self.configuration.items():
            instance_value = getattr(instance, k, None)
            if not instance_value:
                setattr(instance, k, v)

            if isinstance(instance_value, list) and isinstance(v, list):
                setattr(instance, k, instance_value + v)
            elif isinstance(instance_value, dict) and isinstance(v, dict):
                setattr(instance, k, merge_dicts(instance_value, v))
            else:
                setattr(instance, k, v)

    async def wait_stopped(self) -> None:
        if not self._close_waiter:
            self._close_waiter = asyncio.Future()

        await self._close_waiter

    async def run_until_complete(self) -> None:
        services_started: Set = set()
        invoker_tasks: Set = set()
        setup_coros: Set = set()
        teardown_coros: Set = set()
        interrupt_coros: Set = set()
        initialized_coros: Set = set()
        registered_services: Set = set()

        if not self.started_waiter:
            self.started_waiter = asyncio.Future()

        tomodachi.get_contextvar("service.logger").set("tomodachi.service")

        def logging_context_wrapper(coro: Any, service_name_: str, **logger_context: Any) -> Any:
            async def _wrapper(*args: Any, **kwargs: Any) -> Any:
                if not coro:
                    return

                logger = logging.get_logger("tomodachi.service").bind(service=service_name_)

                if logger_context:
                    logger = logging.get_logger().new(**logger_context)

                logging.bind_logger(logger)

                name = logger._context.get("logger", "tomodachi.service")
                if (
                    name in ("service", "tomodachi.service", "tomodachi.handler")
                    or name.startswith("service.")
                    or name.startswith("tomodachi.service.")
                    or name.startswith("tomodachi.handler.")
                    or name.endswith(".handler")
                ):
                    tomodachi.get_contextvar("service.logger").set(name)

                return await coro(*args, **kwargs)

            return _wrapper

        for _, cls in inspect.getmembers(self.module_import):
            if inspect.isclass(cls):
                if not getattr(cls, CLASS_ATTRIBUTE, False):
                    continue

                try:
                    instance = cls()
                except Exception as e:
                    logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    logging.getLogger("tomodachi.init").error(
                        "failed to initialize instance",
                        module=self.module_name,
                        class_=cls.__name__,
                        file_path=self.file_path,
                    )

                    tomodachi.SERVICE_EXIT_CODE = 1
                    self.stop_service()
                    break

                if not getattr(instance, "context", None):
                    setattr(
                        instance,
                        "context",
                        {
                            i: getattr(instance, i)
                            for i in dir(instance)
                            if not callable(i)
                            and not i.startswith("__")
                            and not isinstance(getattr(instance, i), types.MethodType)
                        },
                    )

                getattr(instance, "context", {})["_service_file_path"] = self.file_path

                try:
                    self.setup_configuration(instance)
                except Exception as e:
                    logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    logging.getLogger("tomodachi.init").error(
                        "failed to setup config",
                        module=self.module_name,
                        class_=cls.__name__,
                        file_path=self.file_path,
                    )

                    tomodachi.SERVICE_EXIT_CODE = 1
                    self.stop_service()
                    break

                context_options = getattr(instance, "context", {}).get("options", {})
                if context_options:
                    try:
                        for key in list(context_options.keys()):
                            if "." in key:
                                key_split = key.split(".")
                                op_lvl = context_options
                                for i, k_lvl in enumerate(key_split):
                                    if i + 1 == len(key_split):
                                        if k_lvl in op_lvl and op_lvl[k_lvl] != context_options[key]:
                                            raise Exception(
                                                'Missmatching options for \'{}\': ({}) "{}" and ({}) "{}" differs'.format(
                                                    key,
                                                    type(context_options[key]).__name__,
                                                    context_options[key],
                                                    type(op_lvl[k_lvl]).__name__,
                                                    op_lvl[k_lvl],
                                                )
                                            )
                                        op_lvl[k_lvl] = context_options[key]
                                        continue
                                    if k_lvl not in op_lvl:
                                        op_lvl[k_lvl] = {}
                                    op_lvl = op_lvl.get(k_lvl)
                    except Exception as e:
                        logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                        logging.getLogger("tomodachi.init").error(
                            "failed to setup options",
                            module=self.module_name,
                            class_=cls.__name__,
                            file_path=self.file_path,
                        )

                        tomodachi.SERVICE_EXIT_CODE = 1
                        self.stop_service()
                        break

                if not getattr(instance, "uuid", None):
                    instance.uuid = str(uuid.uuid4())

                service_name = getattr(instance, "name", getattr(cls, "name", None))

                if not service_name:
                    service_name = ServiceContainer.assign_service_name(instance)
                    if not service_name:
                        continue

                set_service(service_name, instance)

                try:
                    post_init_hooks: List[Callable] = [
                        getattr(instance, func_name)
                        for func_name in reversed(dir(instance))
                        if callable(getattr(instance, func_name))
                        and func_name.startswith("_")
                        and func_name.endswith("__post_init_hook")
                    ]
                    for post_init_hook in post_init_hooks:
                        post_init_hook()
                except Exception as e:
                    logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                    logging.getLogger("tomodachi.init").error("failed to run post init hooks", service=service_name)

                    tomodachi.SERVICE_EXIT_CODE = 1
                    self.stop_service()
                    break

                log_level = getattr(instance, "log_level", None) or getattr(cls, "log_level", None) or "INFO"

                def invoker_function_sorter(m: str) -> int:
                    for i, line in enumerate(inspect.getsourcelines(self.module_import)[0]):
                        if re.match(r"^\s*(async)?\s+def\s+{}\s*([(].*$)?$".format(m), line):
                            return i
                    return -1

                invoker_functions = []
                for name, fn in inspect.getmembers(cls):
                    if inspect.isfunction(fn) and getattr(fn, FUNCTION_ATTRIBUTE, None):
                        setattr(fn, START_ATTRIBUTE, True)  # deprecated
                        invoker_functions.append(name)
                invoker_functions.sort(key=invoker_function_sorter)
                if invoker_functions:
                    invoker_tasks = invoker_tasks | set(
                        [
                            (
                                service_name,
                                asyncio.ensure_future(
                                    logging_context_wrapper(
                                        getattr(instance, name),
                                        service_name,
                                        logger="tomodachi.setup",
                                        wrapped_handler=name,
                                    )(**{INVOKER_TASK_START_KEYWORD: True})
                                ),
                            )
                            for name in invoker_functions
                        ]
                    )
                    services_started.add((service_name, instance, log_level))

                if getattr(instance, "_start_service", None):
                    setup_coros.add(
                        logging_context_wrapper(
                            getattr(instance, "_start_service"),
                            service_name,
                            logger="tomodachi.lifecycle.handler",
                            handler="_start_service",
                            type="tomodachi.event",
                            event_="lifecycle.setup",
                        )
                    )
                    services_started.add((service_name, instance, log_level))

                if getattr(instance, "_started_service", None):
                    services_started.add((service_name, instance, log_level))

        if services_started:
            try:
                for name, instance, log_level in services_started:
                    logging.getLogger("tomodachi.service").info(
                        "initializing service instance", service=name, uuid=instance.uuid
                    )
                    self.logger.info(
                        "starting the service",
                        state="starting",
                        service=name if len(services_started) > 1 else Ellipsis,
                    )
                    logging.getLogger("tomodachi.event").debug("-> lifecycle.setup", event_="lifecycle.setup")

                if setup_coros:
                    results = await asyncio.wait([asyncio.ensure_future(func()) for func in setup_coros if func])
                    exceptions = [v.exception() for v in [value for value in results if value][0] if v.exception()]
                    if exceptions:
                        for exception in exceptions:
                            try:
                                raise cast(Exception, exception)
                            except Exception as e:
                                limit_exception_traceback(e, ("tomodachi.container",))
                                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

                        for name, instance, log_level in services_started:
                            self.logger.warning("failed to start service", state="aborting", service=name)
                            logging.getLogger("tomodachi.event").debug("-> lifecycle.abort", event_="lifecycle.abort")

                        if invoker_tasks:
                            await asyncio.gather(*set([future for _, future in invoker_tasks]))
                            invoker_tasks = set()

                        tomodachi.SERVICE_EXIT_CODE = 1
                        self.stop_service()

                if invoker_tasks:
                    await asyncio.gather(*set([future for _, future in invoker_tasks]))
                    invoker_coros = set(
                        [
                            logging_context_wrapper(
                                future.result(),
                                name,
                                logger="tomodachi.invoker",
                                invoker_module=future.result().__module__,
                                invoker_function=future.result().__qualname__.split(".<locals>", 1)[0],
                            )
                            for name, future in invoker_tasks
                            if future and future.result()
                        ]
                    )
                    results = await asyncio.wait([asyncio.ensure_future(func()) for func in invoker_coros if func])
                    exceptions = [v.exception() for v in [value for value in results if value][0] if v.exception()]
                    if exceptions:
                        for exception in exceptions:
                            try:
                                raise cast(Exception, exception)
                            except Exception as e:
                                limit_exception_traceback(e, ("tomodachi.container",))
                                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

                        for name, instance, log_level in services_started:
                            self.logger.warning("failed to start service", state="aborting", service=name)
                            logging.getLogger("tomodachi.event").debug("-> lifecycle.abort", event_="lifecycle.abort")

                        tomodachi.SERVICE_EXIT_CODE = 1
                        self.stop_service()

                if (
                    self.started_waiter
                    and not self.started_waiter.done()
                    and (not self._close_waiter or not self._close_waiter.done())
                ):
                    for name, instance, log_level in services_started:
                        for registry in getattr(instance, "discovery", []):
                            registered_services.add((name, instance))
                            if getattr(registry, "_register_service", None):
                                await asyncio.create_task(
                                    logging_context_wrapper(
                                        registry._register_service,
                                        name,
                                        logger="tomodachi.discovery",
                                        registry=(
                                            registry.name
                                            if hasattr(registry, "name")
                                            else (
                                                registry.__name__
                                                if hasattr(registry, "__name__")
                                                else (
                                                    type(registry).__name__
                                                    if hasattr(type(registry), "__name__")
                                                    else str(type(registry))
                                                )
                                            )
                                        ),
                                        operation="register",
                                    )(instance)
                                )

                        if getattr(instance, "_started_service", None):
                            initialized_coros.add(
                                logging_context_wrapper(
                                    getattr(instance, "_started_service", None),
                                    name,
                                    logger="tomodachi.lifecycle.handler",
                                    handler="_started_service",
                                    type="tomodachi.event",
                                    event_="lifecycle.initialized",
                                )
                            )

                        self.logger.info(
                            "enabled handler functions",
                            state="initialized",
                            service=name if len(services_started) > 1 else Ellipsis,
                        )
                        logging.getLogger("tomodachi.event").debug(
                            "-> lifecycle.initialized", event_="lifecycle.initialized"
                        )

            except Exception as e:
                limit_exception_traceback(e, ("tomodachi.container", "tomodachi.invoker.base"))
                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

                for name, instance, log_level in services_started:
                    self.logger.warning(
                        "failed to start service",
                        state="aborting",
                        service=name if len(services_started) > 1 else Ellipsis,
                    )
                    logging.getLogger("tomodachi.event").debug("-> lifecycle.abort", event_="lifecycle.abort")

                tomodachi.SERVICE_EXIT_CODE = 1
                initialized_coros = set()
                self.stop_service()

            if initialized_coros and any(initialized_coros):
                results = await asyncio.wait([asyncio.ensure_future(func()) for func in initialized_coros if func])
                exceptions = [v.exception() for v in [value for value in results if value][0] if v.exception()]
                if exceptions:
                    for exception in exceptions:
                        try:
                            raise cast(Exception, exception)
                        except Exception as e:
                            limit_exception_traceback(e, ("tomodachi.container",))
                            logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

                    for name, instance, log_level in services_started:
                        self.logger.warning(
                            "failed to start service",
                            state="aborting",
                            service=name if len(services_started) > 1 else Ellipsis,
                        )
                        logging.getLogger("tomodachi.event").debug("-> lifecycle.abort", event_="lifecycle.abort")

                    tomodachi.SERVICE_EXIT_CODE = 1
                    self.stop_service()

        elif not tomodachi.SERVICE_EXIT_CODE:
            logging.getLogger("tomodachi.init").warning(
                "no transport handlers defined", module=self.module_name, file_path=self.file_path
            )
            tomodachi.SERVICE_EXIT_CODE = 1
            self.stop_service()
        else:
            logging.getLogger("tomodachi.init").warning(
                "error during initializing", module=self.module_name, file_path=self.file_path
            )
            self.stop_service()

        self.services_started = services_started
        if self.started_waiter and not self.started_waiter.done():
            self.started_waiter.set_result(services_started)

            if not self._close_waiter or not self._close_waiter.done():
                for name, instance, log_level in services_started:
                    self.logger.info(
                        "started service successfully",
                        state="ready",
                        service=name if len(services_started) > 1 else Ellipsis,
                    )
                    logging.getLogger("tomodachi.event").debug("-> lifecycle.ready", event_="lifecycle.ready")

        await self.wait_stopped()

        for name, instance, log_level in services_started:
            if getattr(instance, "_stopping_service", None):
                interrupt_coros.add(
                    logging_context_wrapper(
                        getattr(instance, "_stopping_service", None),
                        name,
                        logger="tomodachi.lifecycle.handler",
                        handler="_stopping_service",
                        type="tomodachi.event",
                        event_="lifecycle.interrupt",
                    )
                )

            if getattr(instance, "_stop_service", None):
                teardown_coros.add(
                    logging_context_wrapper(
                        getattr(instance, "_stop_service", None),
                        name,
                        logger="tomodachi.lifecycle.handler",
                        handler="_stop_service",
                        type="tomodachi.event",
                        event_="lifecycle.teardown",
                    )
                )

        for name, instance, log_level in services_started:
            logging.getLogger("tomodachi.event").debug("-> lifecycle.interrupt", event_="lifecycle.interrupt")
            self.logger.info(
                "stopping service", state="stopping", service=name if len(services_started) > 1 else Ellipsis
            )

        interrupt_futures = []
        if interrupt_coros and any(interrupt_coros):
            interrupt_futures = [asyncio.ensure_future(func()) for func in interrupt_coros if func]
            await asyncio.sleep(0.01)

        for name, instance, log_level in services_started:
            logging.getLogger("tomodachi.event").debug("-> lifecycle.teardown", event_="lifecycle.teardown")

        for name, instance in registered_services:
            for registry in getattr(instance, "discovery", []):
                if getattr(registry, "_deregister_service", None):
                    try:
                        await asyncio.create_task(
                            logging_context_wrapper(
                                registry._deregister_service,
                                name,
                                logger="tomodachi.discovery",
                                registry=(
                                    registry.name
                                    if hasattr(registry, "name")
                                    else (
                                        registry.__name__
                                        if hasattr(registry, "__name__")
                                        else (
                                            type(registry).__name__
                                            if hasattr(type(registry), "__name__")
                                            else str(type(registry))
                                        )
                                    )
                                ),
                                operation="deregister",
                            )(instance)
                        )
                    except Exception as e:
                        limit_exception_traceback(e, ("tomodachi.container",))
                        logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

        teardown_futures = []
        if teardown_coros and any(teardown_coros):
            teardown_futures = [asyncio.ensure_future(func()) for func in teardown_coros if func]

        if teardown_futures or interrupt_futures:
            results = await asyncio.wait(teardown_futures + interrupt_futures)
            exceptions = [v.exception() for v in [value for value in results if value][0] if v.exception()]
            if exceptions:
                for exception in exceptions:
                    try:
                        raise cast(Exception, exception)
                    except Exception as e:
                        limit_exception_traceback(
                            e,
                            (
                                "tomodachi.container",
                                "tomodachi.transport.http",
                                "tomodachi.transport.aws_sns_sqs",
                                "tomodachi.transport.amqp",
                                "tomodachi.transport.schedule",
                            ),
                        )
                        logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))

        for name, instance, log_level in services_started:
            self.logger.info(
                "terminated service", state="terminated", service=name if len(services_started) > 1 else Ellipsis
            )

        # Debug output if TOMODACHI_DEBUG env is set. Shows still running tasks on service termination.
        if os.environ.get("TOMODACHI_DEBUG") and os.environ.get("TOMODACHI_DEBUG") != "0":
            try:
                tasks = [task for task in asyncio.all_tasks()]
                for task in tasks:
                    try:
                        co_filename = (
                            cast(Any, task.get_coro()).cr_code.co_filename
                            if hasattr(task, "get_coro")
                            else cast(Any, task)._coro.cr_code.co_filename
                        )
                        co_name = (
                            cast(Any, task.get_coro()).cr_code.co_name
                            if hasattr(task, "get_coro")
                            else cast(Any, task).cr_code.co_name
                        )

                        if "/tomodachi/watcher.py" in co_filename and co_name == "_watch_loop":
                            continue
                        if "/tomodachi/container.py" in co_filename and co_name == "run_until_complete":
                            continue
                        if "/asyncio/tasks.py" in co_filename and co_name == "wait":
                            continue

                        self.logger.warning(
                            "task has not been awaited", task_function_name=co_name, task_filename=co_filename
                        )

                    except Exception:
                        pass
            except Exception:
                pass

        for name, instance, log_level in services_started:
            try:
                post_teardown_hooks: List[Callable] = [
                    getattr(instance, func_name)
                    for func_name in reversed(dir(instance))
                    if callable(getattr(instance, func_name))
                    and func_name.startswith("_")
                    and func_name.endswith("__post_teardown_hook")
                ]
                for post_teardown_hook in post_teardown_hooks:
                    post_teardown_hook()
            except Exception as e:
                logging.getLogger("exception").exception("uncaught exception: {}".format(str(e)))
                logging.getLogger("tomodachi.teardown").error(
                    "failed to run post teardown hooks",
                    service=name,
                )

    @classmethod
    def assign_service_name(cls, instance: Any) -> str:
        new_service_name = ""
        if instance.__class__.__module__ and instance.__class__.__module__ not in (
            "service.app",
            "service.service",
            "services.app",
            "services.service",
            "src.service",
            "src.app",
            "code.service",
            "code.app",
            "app.service",
            "app.app",
            "apps.service",
            "apps.app",
            "example.service",
            "example.app",
            "examples.service",
            "examples.app",
            "test.service",
            "test.app",
            "tests.service",
            "tests.app",
        ):
            new_service_name = (
                "{}-".format(
                    re.sub(
                        r"^.*[.]([a-zA-Z0-9_]+)[.]([a-zA-Z0-9_]+)$",
                        r"\1-\2",
                        str(instance.__class__.__module__),
                    )
                )
                .replace("_", "-")
                .replace(".", "-")
            )
        class_name = instance.__class__.__name__
        for i, c in enumerate(class_name.lower()):
            if i and c != class_name[i]:
                new_service_name += "-"
            if c == "_":
                c = "-"
            new_service_name += c

        if new_service_name in ("app", "service"):
            new_service_name = "service"

        if not tomodachi.get_service(new_service_name) and not tomodachi.get_service(
            "{}-0001".format(new_service_name)
        ):
            service_name = new_service_name
        else:
            if tomodachi.get_service(new_service_name) and not tomodachi.get_service(
                "{}-0001".format(new_service_name)
            ):
                other_service = tomodachi.get_service(new_service_name)
                setattr(other_service, "name", "{}-0001".format(new_service_name))
                setattr(other_service.__class__, "name", other_service.name)
                unset_service(new_service_name)
                set_service(other_service.name, other_service)

            incr = 1
            while True:
                test_service_name = "{}-{:04d}".format(new_service_name, incr)
                if tomodachi.get_service(test_service_name):
                    incr += 1
                    continue
                service_name = test_service_name
                break

        setattr(instance, "name", service_name)
        setattr(instance.__class__, "name", service_name)

        return service_name



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from random import randint
from time import time_ns
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from opentelemetry.metrics import Instrument
from opentelemetry.sdk.metrics._internal.aggregation import (
    Aggregation,
    _Aggregation,
    _DropAggregation,
    _ExplicitBucketHistogramAggregation,
)
from opentelemetry.sdk.metrics._internal.instrument import Histogram
from opentelemetry.sdk.metrics._internal.measurement import Measurement
from opentelemetry.sdk.util.instrumentation import InstrumentationScope
from opentelemetry.trace import format_span_id, format_trace_id, get_current_span
from opentelemetry.util.types import Attributes, AttributeValue

from tomodachi.opentelemetry.environment_variables import OTEL_PYTHON_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED

# currently experimental workaround to provide exemplars to metrics in Prometheus OpenMetrics format.
# exemplar spec: https://opentelemetry.io/docs/specs/otel/metrics/sdk/#exemplar

# status of exemplars support for Python OTEL: https://github.com/open-telemetry/opentelemetry-python/issues/2407


def IS_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED() -> bool:
    return str(
        os.environ.get(OTEL_PYTHON_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED)
        or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED")
        or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_EXEMPLARS_ENABLED")
        or os.environ.get("OTEL_EXPORTER_PROMETHEUS_EXEMPLARS_ENABLED")
        or os.environ.get("IS_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED")
        or os.environ.get("TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED")
        or 0
    ).lower().strip() in ("1", "true")


@dataclass(frozen=True)
class Exemplar:
    value: Union[int, float]
    attributes: Attributes
    time_unix_nano: int
    _trace_id: int
    _span_id: int

    @property
    def trace_id(self) -> str:
        return format_trace_id(self._trace_id) or ""

    @property
    def span_id(self) -> str:
        return format_span_id(self._span_id) or ""


T = TypeVar("T", bound="BaseExemplarReservoir")


class BaseExemplarReservoir(ABC):
    _instances: Optional[
        Dict[
            Tuple[InstrumentationScope, str, frozenset[Tuple[str, AttributeValue]]],
            Union[SimpleFixedSizeExemplarReservoir, AlignedHistogramBucketExemplarReservoir],
        ]
    ] = None

    def __new__(
        cls: Type[BaseExemplarReservoir],
        *,
        measurement: Optional[Measurement] = None,
        instrument: Optional[Instrument] = None,
        instrumentation_scope: Optional[InstrumentationScope] = None,
        instrument_name: Optional[str] = None,
        aggregation: Optional[_Aggregation] = None,
        attributes: Optional[Attributes] = None,
    ) -> ExemplarReservoir:
        if cls._instances is None:
            cls._instances = {}

        if measurement:
            instrument = measurement.instrument
            attributes = measurement.attributes
        elif aggregation:
            attributes = aggregation._attributes

        if instrument and not instrumentation_scope:
            instrumentation_scope = getattr(instrument, "instrumentation_scope", None)

        if instrument and not instrument_name:
            instrument_name = getattr(instrument, "name", None) or ""

        if not instrument_name or instrumentation_scope is None:
            raise Exception("instrument or instrument_name + instrumentation_scope is required to load a reservoir")

        aggr_key = frozenset(attributes.items() if attributes is not None else {})
        key = (instrumentation_scope, instrument_name, aggr_key)

        instance: Optional[Union[SimpleFixedSizeExemplarReservoir, AlignedHistogramBucketExemplarReservoir]]
        instance = cls._instances.get(key)

        if not instance:
            if not instrument or not aggregation:
                raise Exception("instrument and aggregation is required to initialize a new reservoir")

            if isinstance(instrument, Histogram):
                instance = object.__new__(AlignedHistogramBucketExemplarReservoir)
            else:
                instance = object.__new__(SimpleFixedSizeExemplarReservoir)

            cls._instances[key] = instance
            instance._initialize(instrument, aggregation, attributes)

        return instance

    @abstractmethod
    def _initialize(self, instrument: Instrument, aggregation: _Aggregation, attributes: Attributes) -> None:
        pass


class ExemplarPool:
    max_exemplars_in_pool: int = 30
    _exemplars: List[Exemplar]

    def __init__(self) -> None:
        self._exemplars = []

    def add_exemplar(self, exemplar: Exemplar) -> None:
        if len(self._exemplars) >= self.max_exemplars_in_pool:
            self._exemplars.pop(0)
        self._exemplars.append(exemplar)

    def clear(self) -> None:
        self._exemplars.clear()


class ExemplarReservoir(BaseExemplarReservoir):
    instrument: Instrument
    aggregation: _Aggregation
    attributes: Mapping[str, AttributeValue]
    _attributes: frozenset[Tuple[str, AttributeValue]]
    _buckets: Tuple[ExemplarPool, ...]

    def _initialize(self, instrument: Instrument, aggregation: _Aggregation, attributes: Attributes) -> None:
        self.instrument = instrument
        self.aggregation = aggregation
        self.attributes = attributes or {}
        self._attributes_set = frozenset(self.attributes.items())
        self._buckets = tuple(ExemplarPool() for _ in range(self.num_buckets))

    def _offer(self, measurement: Measurement) -> Tuple[bool, int]:
        return False, 0

    def _exemplars_from_buckets(self, buckets: Iterable[ExemplarPool]) -> List[Exemplar]:
        exemplars: List[Exemplar] = []
        for bucket in buckets:
            exemplars = exemplars + bucket._exemplars
        return exemplars

    def offer(self, measurement: Measurement, time_unix_nano: int) -> None:
        span_context = get_current_span().get_span_context()

        if span_context.is_valid and span_context.trace_flags.sampled:
            may_sample_exemplar, bucket = self._offer(measurement)
            if may_sample_exemplar and bucket < self.num_buckets:
                self._buckets[bucket].add_exemplar(
                    Exemplar(
                        value=measurement.value,
                        attributes=dict((measurement.attributes or {}).items() - self._attributes_set),
                        time_unix_nano=time_unix_nano,
                        _trace_id=span_context.trace_id,
                        _span_id=span_context.span_id,
                    )
                )

    def collect(
        self,
        filter_key: Callable[[Exemplar], bool],
        sort_key: Callable[[Exemplar], Any],
        bucket: Optional[int] = None,
        limit: int = 1,
    ) -> Iterator[Exemplar]:
        return (
            exemplar
            for exemplar, _ in zip(
                sorted(
                    filter(
                        filter_key,
                        (
                            self._buckets[bucket]._exemplars
                            if bucket is not None
                            else self._exemplars_from_buckets(self._buckets)
                        ),
                    ),
                    key=sort_key,
                ),
                range(limit),
            )
        )

    def reset(self) -> None:
        pass

    @property
    def num_buckets(self) -> int:
        return 0


class SimpleFixedSizeExemplarReservoir(ExemplarReservoir):
    _num_measurements_seen: int

    def _initialize(self, instrument: Instrument, aggregation: _Aggregation, attributes: Attributes) -> None:
        self._num_measurements_seen = 0
        self._num_buckets = 1
        super()._initialize(instrument, aggregation, attributes)

    def _offer(self, measurement: Measurement) -> Tuple[bool, int]:
        self._num_measurements_seen += 1
        return (randint(0, self._num_measurements_seen - 1) == 0, 0)

    def reset(self) -> None:
        self._num_measurements_seen = 0
        self._buckets[0].clear()

    @property
    def num_buckets(self) -> int:
        return 1


class AlignedHistogramBucketExemplarReservoir(ExemplarReservoir):
    _bucket_boundaries: Sequence[float]
    _num_buckets: int

    def _initialize(self, instrument: Instrument, aggregation: _Aggregation, attributes: Attributes) -> None:
        if isinstance(aggregation, _ExplicitBucketHistogramAggregation):
            self._bucket_boundaries = aggregation._boundaries
        else:
            self._bucket_boundaries = getattr(aggregation, "_boundaries", [])
        self._num_buckets = len(self._bucket_boundaries) + 1
        super()._initialize(instrument, aggregation, attributes)

    def _offer(self, measurement: Measurement) -> Tuple[bool, int]:
        return True, self.find_histogram_bucket(measurement)

    def find_histogram_bucket(self, measurement: Measurement) -> int:
        for idx, boundary in enumerate(self._bucket_boundaries):
            if measurement.value <= boundary:
                return idx
        return len(self._bucket_boundaries)

    def reset(self) -> None:
        for bucket in self._buckets:
            if bucket._exemplars:
                exemplar = bucket._exemplars.pop()
                bucket.clear()
                bucket.add_exemplar(exemplar)

    @property
    def num_buckets(self) -> int:
        return self._num_buckets


class _ExemplarAggregation(_DropAggregation):
    def aggregate(self, measurement: Measurement) -> None:
        reservoir = ExemplarReservoir(measurement=measurement)
        reservoir.offer(measurement, time_ns())


class ExemplarAggregation(Aggregation):
    def _create_aggregation(
        self,
        instrument: Instrument,
        attributes: Attributes,
        start_time_unix_nano: int,
    ) -> _Aggregation:
        return _ExemplarAggregation(attributes)



import sys
from typing import Any

from tomodachi._importer import _install_import_finder

_install_import_finder({"tomodachi.transport.awssnssqs": "tomodachi.transport.aws_sns_sqs"})


def __getattr__(name: str) -> Any:
    fullname = f"tomodachi.transport.{name}"
    if fullname in sys.modules and sys.modules[fullname]:
        if fullname == sys.modules[fullname].__name__:
            return sys.modules[fullname]

        fullname = sys.modules[fullname].__name__
        if fullname in sys.modules:
            return sys.modules[fullname]

        name = fullname.rpartition(".")[2]

    import importlib  # noqa  # isort:skip

    try:
        module = importlib.import_module(f"tomodachi.transport.{name}")
    except ModuleNotFoundError:
        raise ImportError(f"cannot import name '{name}' from 'tomodachi.transport' ({__file__})") from None

    return module


__all__ = ["amqp", "aws_sns_sqs", "awssnssqs", "http", "schedule"]



import datetime
from typing import Optional

from tomodachi.__version__ import __build_time__ as tomodachi_build_time


def get_time_since_build(timestamp: Optional[str] = None, build_time: Optional[str] = None) -> str:
    time_since_tomodachi_build = ""

    if not build_time:
        build_time = tomodachi_build_time

    if not timestamp:
        timestamp = (
            datetime.datetime.now(datetime.timezone.utc).isoformat(timespec="microseconds").replace("+00:00", "Z")
        )

    if build_time:
        try:
            tomodachi_build_datetime = datetime.datetime.strptime(build_time, "%Y-%m-%dT%H:%M:%S.%f%z").astimezone(
                datetime.timezone.utc
            )
            current_datetime = datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f%z").astimezone(
                datetime.timezone.utc
            )
            timedelta_since_tomodachi_build = current_datetime - tomodachi_build_datetime
            if timedelta_since_tomodachi_build.days == 0:
                seconds = timedelta_since_tomodachi_build.seconds
                if seconds >= 3600:
                    hours = seconds // 3600
                    time_since_tomodachi_build = f"released {hours} hour{'s' if hours > 1 else ''} ago"
                elif seconds >= 60:
                    minutes = seconds // 60
                    time_since_tomodachi_build = f"released {minutes} minute{'s' if minutes > 1 else ''} ago"
                else:
                    time_since_tomodachi_build = "released just now"
            else:
                days = timedelta_since_tomodachi_build.days
                if days >= 720:
                    years = days // 365
                    if years < 2:
                        years = 2
                    time_since_tomodachi_build = f"released over {years} year{'s' if years > 1 else ''} ago"
                elif days >= 90:
                    months = days // 30
                    time_since_tomodachi_build = f"released {months} month{'s' if months > 1 else ''} ago"
                else:
                    time_since_tomodachi_build = f"released {days} day{'s' if days > 1 else ''} ago"
        except Exception:
            pass

    return time_since_tomodachi_build



from os import environ
from typing import List, Union

from opentelemetry.instrumentation.auto_instrumentation._load import (
    _load_configurators,
    _load_distro,
    _load_instrumentors,
)
from opentelemetry.instrumentation.environment_variables import (
    OTEL_PYTHON_CONFIGURATOR,
    OTEL_PYTHON_DISABLED_INSTRUMENTATIONS,
    OTEL_PYTHON_DISTRO,
)
from pkg_resources import Distribution, EntryPoint, iter_entry_points

from tomodachi.opentelemetry.distro import OpenTelemetryConfigurator, OpenTelemetryDistro


def initialize() -> None:
    # use default "tomodachi" distro and configurator unless otherwise specified
    if environ.get(OTEL_PYTHON_DISTRO) is None:
        environ.setdefault(OTEL_PYTHON_DISTRO, "tomodachi")
    if environ.get(OTEL_PYTHON_CONFIGURATOR) is None:
        environ.setdefault(OTEL_PYTHON_CONFIGURATOR, "tomodachi")
    distro_value = environ.get(OTEL_PYTHON_DISTRO) or "tomodachi"
    configurator_value = environ.get(OTEL_PYTHON_CONFIGURATOR) or "tomodachi"

    # load "tomodachi" distro directly unless otherwise specified (then use standard OTEL auto loader)
    if distro_value != "tomodachi":
        distro = _load_distro()
    else:
        distro = OpenTelemetryDistro()
    distro.configure()

    # load "tomodachi" configurator directly unless otherwise specified (then use standard OTEL auto loader)
    if configurator_value != "tomodachi":
        _load_configurators()
    else:
        OpenTelemetryConfigurator().configure()

    # loads instrumentors from entry points using selected distro
    _load_instrumentors(distro)

    # auto instrument also in case "tomodachi" instrumentor plugin is missing (or tomodachi dist is not installed)
    if "tomodachi" not in [ep.name for ep in iter_entry_points("opentelemetry_instrumentor")]:
        package_to_exclude: Union[str, List[str]] = environ.get(OTEL_PYTHON_DISABLED_INSTRUMENTATIONS, [])
        if isinstance(package_to_exclude, str):
            package_to_exclude = [x.strip() for x in package_to_exclude.split(",")]

        if "tomodachi" not in package_to_exclude:
            if isinstance(distro, OpenTelemetryDistro):
                from tomodachi.opentelemetry import TomodachiInstrumentor

                if TomodachiInstrumentor not in distro._instrumentors:
                    TomodachiInstrumentor().instrument(skip_dep_check=True)
                    distro._instrumentors.add(TomodachiInstrumentor)
            else:
                distro.load_instrumentor(
                    EntryPoint("tomodachi", "tomodachi.opentelemetry", ("TomodachiInstrumentor",), dist=Distribution()),
                    skip_dep_check=True,
                )



import os
import re
from collections import deque
from json import dumps
from re import IGNORECASE, UNICODE, compile
from typing import Any, Callable, Deque, Dict, Generator, List, Optional, cast

from opentelemetry.exporter.prometheus import _CustomCollector as _PrometheusCustomCollector
from opentelemetry.sdk.metrics import Meter, MeterProvider
from opentelemetry.sdk.metrics._internal.point import HistogramDataPoint, NumberDataPoint
from opentelemetry.sdk.metrics.export import Metric, MetricReader, MetricsData, ResourceMetrics, ScopeMetrics
from opentelemetry.sdk.resources import SERVICE_INSTANCE_ID as RESOURCE_SERVICE_INSTANCE_ID
from opentelemetry.sdk.resources import SERVICE_NAME as RESOURCE_SERVICE_NAME
from opentelemetry.sdk.resources import SERVICE_NAMESPACE as RESOURCE_SERVICE_NAMESPACE
from opentelemetry.util.types import AttributeValue
from prometheus_client import Info
from prometheus_client.core import Metric as PrometheusMetric
from prometheus_client.registry import Collector, CollectorRegistry
from prometheus_client.samples import Exemplar as PrometheusExemplar
from prometheus_client.samples import Sample as PrometheusSample

from tomodachi import logging
from tomodachi.opentelemetry.distro import _create_resource
from tomodachi.opentelemetry.environment_variables import (
    OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS,
    OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO,
    OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO,
    OTEL_PYTHON_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX,
    OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT,
)
from tomodachi.opentelemetry.exemplars import IS_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED, Exemplar, ExemplarReservoir

# currently experimental workaround to provide metrics for Prometheus to scrape according to specification.


UNIT_TRANSFORM_MAP = {
    "s": "seconds",
    "d": "days",
    "h": "hours",
    "min": "minutes",
    "ms": "milliseconds",
    "us": "microseconds",
    "ns": "nanoseconds",
    "second": "seconds",
    "m": "meters",
    "v": "volts",
    "a": "amperes",
    "j": "joules",
    "w": "watts",
    "g": "grams",
    "cel": "celsius",
    "hz": "hertz",
    "request": "requests",
    "req": "requests",
    "task": "tasks",
    "message": "messages",
    "msg": "messages",
    "byte": "bytes",
    "by": "bytes",
    "kiby": "kibibytes",
    "miby": "mebibytes",
    "giby": "gibibytes",
    "tiby": "tebibytes",
    "kby": "kilobytes",
    "mby": "megabytes",
    "gby": "gigabytes",
    "tby": "terabytes",
    "page": "pages",
    "fault": "faults",
    "query": "queries",
    "invocation": "invocations",
    "operation": "operations",
    "packet": "packets",
    "error": "errors",
    "connection": "connections",
    "thread": "threads",
    "class": "classes",
    "buffer": "buffers",
    "%": "percent",
    "1": "ratio",
}

PER_UNIT_TRANSFORM_MAP = {
    "s": "second",
    "m": "minute",
    "h": "hour",
    "d": "day",
    "w": "week",
    "mo": "month",
    "y": "year",
}

REGISTRY_ = CollectorRegistry(auto_describe=True)


class _CustomCollector(_PrometheusCustomCollector):
    _metrics_datas: Deque[MetricsData]
    _callback: Callable  # type: ignore

    def __init__(self, prefix: str = "", registry: CollectorRegistry = REGISTRY_) -> None:
        super().__init__()
        self._prefix = prefix
        self._registry = registry
        self._exemplars: Deque[List[Optional[Exemplar]]] = deque()
        self._include_scope_info = bool(
            str(
                os.environ.get(OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO)
                or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_INCLUDE_SCOPE_INFO")
                or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_INCLUDE_OTEL_SCOPE_INFO")
                or os.environ.get("OTEL_EXPORTER_PROMETHEUS_INCLUDE_OTEL_SCOPE_INFO")
                or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_INCLUDE_SCOPE_INFO")
                or os.environ.get("OTEL_EXPORTER_PROMETHEUS_INCLUDE_SCOPE_INFO")
                or 1
            )
            .lower()
            .strip()
            in (
                "1",
                "true",
            )
        )

    def add_exemplars(self, exemplars_data: List[Optional[Exemplar]]) -> None:
        self._exemplars.append(exemplars_data)

    def _is_valid_exemplar_metric(self, metric: PrometheusMetric, sample: PrometheusSample) -> bool:
        if metric.type == "counter" and sample.name.endswith("_total"):
            return True
        if metric.type in ("histogram", "gaugehistogram") and sample.name.endswith("_bucket"):
            return True
        return False

    def collect(self) -> Generator[PrometheusMetric, Any, Any]:  # type: ignore
        if self._callback is not None:
            self._callback()

        if self._metrics_datas:
            if self._include_scope_info:
                otel_scope_info_collector = Info(
                    "otel_scope",
                    "Instrumentation Scope metadata",
                    labelnames=("otel_scope_name", "otel_scope_version"),
                    registry=None,
                )
                for metrics_data in self._metrics_datas:
                    for resource_metric in metrics_data.resource_metrics:
                        for scope_metric in resource_metric.scope_metrics:
                            otel_scope_info_collector.labels(
                                otel_scope_name=scope_metric.scope.name,
                                otel_scope_version=scope_metric.scope.version or "",
                            )
                for metric_family in otel_scope_info_collector.collect():
                    yield metric_family

            target_info: Dict[str, str] = self._registry.get_target_info() or {}
            add_metric_labels = {k: v for k, v in target_info.items() if k in ("job", "instance")}

            metric_family_id_metric_family: Dict[str, PrometheusMetric] = {}

            while self._metrics_datas:
                metrics_data = self._metrics_datas.popleft()
                add_metric_labels_ = add_metric_labels.copy()
                if self._include_scope_info:
                    for resource_metric in metrics_data.resource_metrics:
                        for scope_metric in resource_metric.scope_metrics:
                            add_metric_labels_["otel_scope_name"] = scope_metric.scope.name
                            add_metric_labels_["otel_scope_version"] = scope_metric.scope.version or ""

                self.translate_to_prometheus(
                    metrics_data, self._exemplars.popleft(), add_metric_labels_, metric_family_id_metric_family
                )

                if metric_family_id_metric_family:
                    for metric_family in metric_family_id_metric_family.values():
                        yield metric_family

    def translate_to_prometheus(
        self,
        metrics_data: MetricsData,
        exemplars: List[Optional[Exemplar]],
        add_metric_labels: Dict[str, str],
        metric_family_id_metric_family: Dict[str, PrometheusMetric],
    ) -> None:
        metric_family_id_metric_family.clear()
        self._translate_to_prometheus(metrics_data, metric_family_id_metric_family)
        if metric_family_id_metric_family:
            if add_metric_labels:
                for metric_family in metric_family_id_metric_family.values():
                    for sample in metric_family.samples:
                        sample.labels.update(add_metric_labels)

            for metric_family in metric_family_id_metric_family.values():
                if metric_family.type in ("histogram", "gaugehistogram"):
                    for idx, sample in enumerate(metric_family.samples[:]):
                        if not self._is_valid_exemplar_metric(metric_family, sample):
                            continue
                        exemplar = exemplars.pop(0) if exemplars else None
                        if exemplar and not sample.exemplar:
                            exemplar_labels = {
                                **{
                                    self._sanitize(k): self._check_value(v)
                                    for k, v in (exemplar.attributes or {}).items()
                                },
                                "trace_id": exemplar.trace_id,
                                "span_id": exemplar.span_id,
                            }

                            metric_family.samples[idx] = sample._replace(
                                exemplar=PrometheusExemplar(
                                    exemplar_labels,
                                    exemplar.value,
                                    exemplar.time_unix_nano / 1e9,
                                )
                            )


class TomodachiPrometheusMetricReader(MetricReader):
    def __init__(self, prefix: str = "", registry: CollectorRegistry = REGISTRY_) -> None:
        super().__init__()
        self._collector = _CustomCollector(prefix, registry)
        self._registry = registry
        self._registry.register(cast(Collector, self._collector))
        setattr(self._collector, "_callback", self.collect)

    def shutdown(self, timeout_millis: float = 30_000, **kwargs: Any) -> None:
        super().shutdown(timeout_millis=timeout_millis, **kwargs)
        self._registry.unregister(cast(Collector, self._collector))

    def _transform_metric(self, metric: Metric) -> Metric:
        # https://github.com/open-telemetry/opentelemetry-java/issues/5529
        # https://github.com/open-telemetry/opentelemetry-collector-contrib/blob/main/pkg/translator/prometheus/normalize_name.go#L19

        unit = str(metric.unit.replace("{", "").replace("}", "") if metric.unit else "").lower()
        if "/" in unit:
            u, per_u = unit.split("/", 1)
            u = UNIT_TRANSFORM_MAP.get(u, u)
            per_u = PER_UNIT_TRANSFORM_MAP.get(per_u, per_u)
        else:
            unit = UNIT_TRANSFORM_MAP.get(unit, unit)

        metric_name = metric.name if not self._collector._prefix else f"{self._collector._prefix}_{metric.name}"

        return Metric(
            name=metric_name,
            description=metric.description,
            unit=unit,
            data=metric.data,
        )

    def _receive_metrics(
        self,
        metrics_data: MetricsData,
        timeout_millis: float = 10_000,
        **kwargs: Any,
    ) -> None:
        if metrics_data is None or not metrics_data.resource_metrics:
            return

        for resource_metric in metrics_data.resource_metrics:
            for scope_metric in resource_metric.scope_metrics:
                for metric in scope_metric.metrics:
                    metrics_data_ = MetricsData(
                        resource_metrics=[
                            ResourceMetrics(
                                resource=resource_metric.resource,
                                scope_metrics=[
                                    ScopeMetrics(
                                        scope=scope_metric.scope,
                                        metrics=[self._transform_metric(metric)],
                                        schema_url=scope_metric.schema_url,
                                    )
                                ],
                                schema_url=resource_metric.schema_url,
                            )
                        ]
                    )
                    self._collector.add_metrics_data(metrics_data_)

                    if not IS_TOMODACHI_PROMETHEUS_EXEMPLARS_ENABLED():
                        self._collector.add_exemplars([])
                        continue

                    # currently only provides exemplars for metrics on the "tomodachi" scope
                    # todo: add support to add exemplars on any instrumentation scope
                    if scope_metric.scope.name != "tomodachi":
                        self._collector.add_exemplars([])
                        continue

                    exemplars: List[Optional[Exemplar]] = []
                    for data_point in metric.data.data_points:
                        reservoir = ExemplarReservoir(
                            instrument_name=metric.name,
                            instrumentation_scope=scope_metric.scope,
                            attributes=data_point.attributes,
                        )

                        if isinstance(data_point, HistogramDataPoint):
                            current_exemplar = None
                            for bucket_count, explicit_bounds, bucket in zip(
                                data_point.bucket_counts,
                                (*data_point.explicit_bounds, None),
                                range(len(data_point.bucket_counts)),
                            ):
                                if bucket_count:
                                    for exemplar in reservoir.collect(
                                        filter_key=lambda e: e.time_unix_nano >= data_point.start_time_unix_nano
                                        and e.time_unix_nano <= data_point.time_unix_nano
                                        and (explicit_bounds is None or e.value <= explicit_bounds),
                                        sort_key=lambda e: -e.value,
                                        bucket=bucket,
                                        limit=1,
                                    ):
                                        current_exemplar = exemplar
                                exemplars.append(current_exemplar)
                        elif isinstance(data_point, NumberDataPoint):
                            is_monotonic = getattr(reservoir.aggregation, "_instrument_is_monotonic", None)
                            current_exemplar = None
                            if is_monotonic is None:
                                is_monotonic = getattr(metric.data, "is_monotonic", None)
                            if is_monotonic:
                                for exemplar in reservoir.collect(
                                    filter_key=lambda e: bool(
                                        e.time_unix_nano >= data_point.start_time_unix_nano
                                        and e.time_unix_nano <= data_point.time_unix_nano
                                    ),
                                    sort_key=lambda e: -e.time_unix_nano,
                                    limit=1,
                                ):
                                    current_exemplar = exemplar
                            exemplars.append(current_exemplar)

                        reservoir.reset()

                    self._collector.add_exemplars(exemplars)


class TomodachiPrometheusMeterProvider(MeterProvider):
    def __init__(self, prefix: Optional[str] = None, registry: CollectorRegistry = REGISTRY_) -> None:
        self._prometheus_server_started = False
        self._prometheus_registry = registry
        self._non_letters_digits_underscore_re = compile(r"[^\w]", UNICODE | IGNORECASE)

        if prefix is None:
            prefix = str(
                os.environ.get(OTEL_PYTHON_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX)
                or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_NAMESPACE_PREFIX")
                or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_NAMESPACE_PREFIX")
                or os.environ.get("OTEL_EXPORTER_PROMETHEUS_NAMESPACE_PREFIX")
                or ""
            )

        resource = _create_resource()
        reader = TomodachiPrometheusMetricReader(prefix, registry)

        super().__init__(metric_readers=[reader], resource=resource)

    def _get_target_info(self) -> Dict[str, str]:
        # https://opentelemetry.io/docs/specs/otel/compatibility/prometheus_and_openmetrics/#resource-attributes-1
        # https://github.com/OpenObservability/OpenMetrics/blob/main/specification/OpenMetrics.md#supporting-target-metadata-in-both-push-based-and-pull-based-systems
        # https://github.com/open-telemetry/opentelemetry-collector-contrib/blob/main/exporter/prometheusexporter/collector.go#L112-L137
        resource = self._sdk_config.resource

        service_name = str(resource.attributes.get(RESOURCE_SERVICE_NAME) or "")
        service_namespace = str(resource.attributes.get(RESOURCE_SERVICE_NAMESPACE) or "")
        service_instance_id = str(resource.attributes.get(RESOURCE_SERVICE_INSTANCE_ID) or "")

        job = f"{service_namespace}/{service_name}" if service_namespace else service_name
        instance = service_instance_id

        def _sanitize_key(key: str) -> str:
            return self._non_letters_digits_underscore_re.sub("_", key)

        def _sanitize_value(value: AttributeValue) -> str:
            return str(value if isinstance(value, str) else dumps(value, default=str))

        resource_attributes = {_sanitize_key(k): _sanitize_value(v) for k, v in resource.attributes.items()}
        target_info: Dict[str, str] = self._prometheus_registry.get_target_info() or {}
        return {**resource_attributes, **target_info, "job": job, "instance": instance}

    def _start_prometheus_http_server(self) -> None:
        if self._prometheus_server_started:
            return

        import prometheus_client

        addr = str(
            os.environ.get(OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS)
            or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_METER_PROVIDER_ADDRESS")
            or os.environ.get("OTEL_PYTHON_TOMODACHI_PROMETHEUS_ADDRESS")
            or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_ADDRESS")
            or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_HOST")
            or os.environ.get("OTEL_EXPORTER_PROMETHEUS_HOST")
            or "localhost"
        )
        port = int(
            os.environ.get(OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT)
            or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_METER_PROVIDER_PORT")
            or os.environ.get("OTEL_PYTHON_TOMODACHI_PROMETHEUS_PORT")
            or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_PORT")
            or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_PORT")
            or os.environ.get("OTEL_EXPORTER_PROMETHEUS_PORT")
            or 9464
        )

        if str(
            os.environ.get(OTEL_PYTHON_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO)
            or os.environ.get("OTEL_TOMODACHI_PROMETHEUS_INCLUDE_TARGET_INFO")
            or os.environ.get("OTEL_PYTHON_EXPORTER_PROMETHEUS_INCLUDE_TARGET_INFO")
            or os.environ.get("OTEL_EXPORTER_PROMETHEUS_INCLUDE_TARGET_INFO")
            or 1
        ).lower().strip() in (
            "1",
            "true",
        ):
            try:
                self._prometheus_registry.set_target_info(self._get_target_info())
            except Exception as e:
                logging.get_logger("tomodachi.opentelemetry").warning(
                    "unable to add target_info to prometheus collector registry",
                    error=str(e),
                    opentelemetry_metric_provider="tomodachi_prometheus",
                )

        try:
            prometheus_client.start_http_server(port=port, addr=addr, registry=self._prometheus_registry)
        except OSError as e:
            error_message = re.sub(".*: ", "", e.strerror)
            logging.get_logger("tomodachi.opentelemetry").warning(
                "unable to bind prometheus http server [http] to http://{}:{}/ in otel metric provider".format(
                    "localhost" if addr in ("0.0.0.0", "127.0.0.1") else addr, port
                ),
                host=addr,
                port=port,
                error_message=error_message,
                opentelemetry_metric_provider="tomodachi_prometheus",
            )

            try:
                raise Exception(str(e)).with_traceback(e.__traceback__) from None
            except Exception as exc:
                exc.__traceback__ = e.__traceback__
                raise

        listen_url = "http://{}:{}/".format("localhost" if addr in ("0.0.0.0", "127.0.0.1") else addr, port)
        logging.get_logger("tomodachi.opentelemetry").info(
            "prometheus client started",
            listen_url=listen_url,
            listen_host=addr,
            listen_port=port,
        )
        self._prometheus_server_started = True

    def get_meter(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
    ) -> Meter:
        self._start_prometheus_http_server()
        return super().get_meter(name, version=version, schema_url=schema_url)



from typing import Any, Dict

__cached_defs: Dict[str, Any] = {}


def __getattr__(name: str) -> Any:
    if name in __cached_defs:
        return __cached_defs[name]

    import importlib  # noqa  # isort:skip

    name_ = name
    result: Any
    if name in ("AWSSNSRegistration", "awssns", "aws_sns"):
        name = "AWSSNSRegistration"
        module = importlib.import_module(".aws_sns_registration", "tomodachi.discovery")
        result = getattr(module, name)
    elif name in ("DummyRegistry", "dummy", "example"):
        name = "DummyRegistry"
        module = importlib.import_module(".dummy_registry", "tomodachi.discovery")
        result = getattr(module, name)
    elif name in ("aws_sns_registration",):
        name = "aws_sns_registration"
        module = importlib.import_module(".aws_sns_registration", "tomodachi.discovery")
        result = module
    elif name in ("dummy_registry",):
        name = "dummy_registry"
        module = importlib.import_module(".dummy_registry", "tomodachi.discovery")
        result = module
    else:
        raise AttributeError("module 'tomodachi.discovery' has no attribute '{}'".format(name))

    __cached_defs[name] = __cached_defs[name_] = result
    return __cached_defs[name]


__all__ = [
    "AWSSNSRegistration",
    "awssns",
    "aws_sns",
    "DummyRegistry",
    "dummy",
    "example",
    "aws_sns_registration",
    "dummy_registry",
]
